from modelx.serialize.jsonvalues import *

_name = "solvency2"

_allow_none = False

_spaces = [
    "Input",
    "LifeTable",
    "Economic",
    "BaseProj",
    "PV",
    "Override",
    "SCR_life",
    "Policy",
    "Assumptions"
]

