from modelx.serialize.jsonvalues import *

_name = "fastlife"

_allow_none = False

_spaces = [
    "Input",
    "LifeTable",
    "Economic",
    "PV",
    "Projection"
]

# ---------------------------------------------------------------------------
# References

pd = ("Module", "pandas")

np = ("Module", "numpy")